@@ -1,5 +1,5 @@
 /* Copyright (C) 1995-1999, 2000, 2001, 2002, 2003, 2004, 2006, 2008,
- *   2009, 2010, 2011, 2012, 2013, 2014, 2015 Free Software Foundation, Inc.
+ *   2009, 2010, 2011, 2012 Free Software Foundation, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -30,7 +30,6 @@
 
 #include <uniconv.h>
 #include <unictype.h>
-#include <c-strcase.h>
 
 #include "libguile/_scm.h"
 #include "libguile/chars.h"
@@ -46,7 +45,6 @@
 #include "libguile/alist.h"
 #include "libguile/struct.h"
 #include "libguile/ports.h"
-#include "libguile/ports-internal.h"
 #include "libguile/root.h"
 #include "libguile/strings.h"
 #include "libguile/strports.h"
@@ -63,15 +61,9 @@
 
 /* Character printers.  */
 
-#define PORT_CONVERSION_HANDLER(port)		\
-  SCM_PTAB_ENTRY (port)->ilseq_handler
-
 static size_t display_string (const void *, int, size_t, SCM,
 			      scm_t_string_failed_conversion_handler);
 
-static size_t write_string (const void *, int, size_t, SCM,
-			    scm_t_string_failed_conversion_handler);
-
 static int display_character (scm_t_wchar, SCM,
 			      scm_t_string_failed_conversion_handler);
 
@@ -118,8 +110,6 @@ scm_t_option scm_print_opts[] = {
     "'reader' quotes them when the reader option 'keywords' is not '#f'." },
   { SCM_OPTION_BOOLEAN, "escape-newlines", 1,
     "Render newlines as \\n when printing using `write'." },
-  { SCM_OPTION_BOOLEAN, "r7rs-symbols", 0,
-    "Escape symbols using R7RS |...| symbol notation." },
   { 0 },
 };
 
@@ -362,10 +352,6 @@ symbol_has_extended_read_syntax (SCM sym)
   /* Other initial-character constraints.  */
   if (c == '\'' || c == '`' || c == ',' || c == '"' || c == ';' || c == '#')
     return 1;
-
-  /* R7RS allows neither '|' nor '\' in bare symbols.  */
-  if ((c == '|' || c == '\\') && SCM_PRINT_R7RS_SYMBOLS_P)
-    return 1;
   
   /* Keywords can be identified by trailing colons too.  */
   if (c == ':' || scm_i_symbol_ref (sym, len - 1) == ':')
@@ -389,9 +375,6 @@ symbol_has_extended_read_syntax (SCM sym)
         return 1;
       else if (c == '"' || c == ';' || c == '#')
         return 1;
-      else if ((c == '|' || c == '\\') && SCM_PRINT_R7RS_SYMBOLS_P)
-        /* R7RS allows neither '|' nor '\' in bare symbols.  */
-        return 1;
     }
 
   return 0;
@@ -410,7 +393,7 @@ print_extended_symbol (SCM sym, SCM port)
   scm_t_string_failed_conversion_handler strategy;
 
   len = scm_i_symbol_length (sym);
-  strategy = PORT_CONVERSION_HANDLER (port);
+  strategy = scm_i_get_conversion_strategy (port);
 
   scm_lfwrite ("#{", 2, port);
 
@@ -422,84 +405,30 @@ print_extended_symbol (SCM sym, SCM port)
                                             SUBSEQUENT_IDENTIFIER_MASK
                                             | UC_CATEGORY_MASK_Zs))
         {
-          if (!display_character (c, port, strategy)
-              || (c == '\\' && !display_character (c, port, strategy)))
+          if (!display_character (c, port, strategy))
             scm_encoding_error ("print_extended_symbol", errno,
                                 "cannot convert to output locale",
                                 port, SCM_MAKE_CHAR (c));
         }
       else
         {
-          scm_lfwrite ("\\x", 2, port);
+          display_string ("\\x", 1, 2, port, iconveh_question_mark);
           scm_intprint (c, 16, port);
-          scm_putc (';', port);
+          display_character (';', port, iconveh_question_mark);
         }
     }
 
   scm_lfwrite ("}#", 2, port);
 }
 
-static void
-print_r7rs_extended_symbol (SCM sym, SCM port)
-{
-  size_t pos, len;
-  scm_t_string_failed_conversion_handler strategy;
-
-  len = scm_i_symbol_length (sym);
-  strategy = PORT_CONVERSION_HANDLER (port);
-
-  scm_putc ('|', port);
-
-  for (pos = 0; pos < len; pos++)
-    {
-      scm_t_wchar c = scm_i_symbol_ref (sym, pos);
-
-      switch (c)
-        {
-        case '\a': scm_lfwrite ("\\a", 2, port); break;
-        case '\b': scm_lfwrite ("\\b", 2, port); break;
-        case '\t': scm_lfwrite ("\\t", 2, port); break;
-        case '\n': scm_lfwrite ("\\n", 2, port); break;
-        case '\r': scm_lfwrite ("\\r", 2, port); break;
-        case '|':  scm_lfwrite ("\\|", 2, port); break;
-        case '\\': scm_lfwrite ("\\x5c;", 5, port); break;
-        default:
-          if (uc_is_general_category_withtable (c,
-                                                UC_CATEGORY_MASK_L
-                                                | UC_CATEGORY_MASK_M
-                                                | UC_CATEGORY_MASK_N
-                                                | UC_CATEGORY_MASK_P
-                                                | UC_CATEGORY_MASK_S)
-              || (c == ' '))
-            {
-              if (!display_character (c, port, strategy))
-                scm_encoding_error ("print_r7rs_extended_symbol", errno,
-                                    "cannot convert to output locale",
-                                    port, SCM_MAKE_CHAR (c));
-            }
-          else
-            {
-              scm_lfwrite ("\\x", 2, port);
-              scm_intprint (c, 16, port);
-              scm_putc (';', port);
-            }
-          break;
-        }
-    }
-
-  scm_putc ('|', port);
-}
-
-/* FIXME: allow R6RS hex escapes instead of #{...}# or |...|.  */
+/* FIXME: allow R6RS hex escapes instead of #{...}#.  */
 void
 scm_i_print_symbol_name (SCM sym, SCM port)
 {
-  if (!symbol_has_extended_read_syntax (sym))
-    print_normal_symbol (sym, port);
-  else if (SCM_PRINT_R7RS_SYMBOLS_P)
-    print_r7rs_extended_symbol (sym, port);
-  else
+  if (symbol_has_extended_read_syntax (sym))
     print_extended_symbol (sym, port);
+  else
+    print_normal_symbol (sym, port);
 }
 
 void
@@ -571,7 +500,7 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 	  else
 	    {
 	      if (!display_character (SCM_CHAR (exp), port,
-				      PORT_CONVERSION_HANDLER (port)))
+				      scm_i_get_conversion_strategy (port)))
 		scm_encoding_error (__func__, errno,
 				    "cannot convert to output locale",
 				    port, exp);
@@ -637,29 +566,32 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
           }
 	  break;
         case scm_tc7_string:
-	  {
-	    size_t len, printed;
+          if (SCM_WRITINGP (pstate))
+            {
+              size_t len, i;
 
-	    len = scm_i_string_length (exp);
-	    if (SCM_WRITINGP (pstate))
-	      {
-		printed = write_string (scm_i_string_data (exp),
-					scm_i_is_narrow_string (exp),
-					len, port,
-					PORT_CONVERSION_HANDLER (port));
-		len += 2;		    /* account for the quotes */
-	      }
-	    else
+              display_character ('"', port, iconveh_question_mark);
+              len = scm_i_string_length (exp);
+              for (i = 0; i < len; ++i)
+		write_character (scm_i_string_ref (exp, i), port, 1);
+
+              display_character ('"', port, iconveh_question_mark);
+              scm_remember_upto_here_1 (exp);
+            }
+          else
+	    {
+	      size_t len, printed;
+
+	      len = scm_i_string_length (exp);
 	      printed = display_string (scm_i_string_data (exp),
 					scm_i_is_narrow_string (exp),
 					len, port,
-					PORT_CONVERSION_HANDLER (port));
-
-	    if (SCM_UNLIKELY (printed < len))
-	      scm_encoding_error (__func__, errno,
-				  "cannot convert to output locale",
-				  port, scm_c_string_ref (exp, printed));
-	  }
+					scm_i_get_conversion_strategy (port));
+	      if (SCM_UNLIKELY (printed < len))
+		scm_encoding_error (__func__, errno,
+				    "cannot convert to output locale",
+				    port, scm_c_string_ref (exp, printed));
+	    }
 
           scm_remember_upto_here_1 (exp);
           break;
@@ -752,7 +684,7 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 		   `SIMPLE_VECTOR_REF ()' macro.  */
 		for (i = 0; i < last; ++i)
 		  {
-		    scm_iprin1 (scm_c_weak_vector_ref (exp, i),
+		    scm_iprin1 (scm_c_vector_ref (exp, i),
 				port, pstate);
 		    scm_putc (' ', port);
 		  }
@@ -769,10 +701,7 @@ iprin1 (SCM exp, SCM port, scm_print_state *pstate)
 	    if (i == last)
 	      {
 		/* CHECK_INTS; */
-                scm_iprin1 (SCM_I_WVECTP (exp)
-                            ? scm_c_weak_vector_ref (exp, i)
-                            : SCM_SIMPLE_VECTOR_REF (exp, i),
-                            port, pstate);
+		scm_iprin1 (scm_c_vector_ref (exp, i), port, pstate);
 	      }
 	    if (cutp)
 	      scm_puts (" ...", port);
@@ -947,25 +876,9 @@ display_string_using_iconv (const void *str, int narrow_p, size_t len,
 			    scm_t_string_failed_conversion_handler strategy)
 {
   size_t printed;
-  scm_t_iconv_descriptors *id;
-  scm_t_port_internal *pti = SCM_PORT_GET_INTERNAL (port);
+  scm_t_port *pt;
 
-  id = scm_i_port_iconv_descriptors (port, SCM_PORT_WRITE);
-
-  if (SCM_UNLIKELY (pti->at_stream_start_for_bom_write && len > 0))
-    {
-      scm_t_port *pt = SCM_PTAB_ENTRY (port);
-
-      /* Record that we're no longer at stream start.  */
-      pti->at_stream_start_for_bom_write = 0;
-      if (pt->rw_random)
-        pti->at_stream_start_for_bom_read = 0;
-
-      /* Write a BOM if appropriate.  */
-      if (SCM_UNLIKELY (c_strcasecmp(pt->encoding, "UTF-16") == 0
-                        || c_strcasecmp(pt->encoding, "UTF-32") == 0))
-        display_character (SCM_UNICODE_BOM, port, iconveh_error);
-    }
+  pt = SCM_PTAB_ENTRY (port);
 
   printed = 0;
 
@@ -994,7 +907,7 @@ display_string_using_iconv (const void *str, int narrow_p, size_t len,
       output = encoded_output;
       output_left = sizeof (encoded_output);
 
-      done = iconv (id->output_cd, &input, &input_left,
+      done = iconv (pt->output_cd, &input, &input_left,
 		    &output, &output_left);
 
       output_len = sizeof (encoded_output) - output_left;
@@ -1004,7 +917,7 @@ display_string_using_iconv (const void *str, int narrow_p, size_t len,
           int errno_save = errno;
 
 	  /* Reset the `iconv' state.  */
-	  iconv (id->output_cd, NULL, NULL, NULL, NULL);
+	  iconv (pt->output_cd, NULL, NULL, NULL, NULL);
 
 	  /* Print the OUTPUT_LEN bytes successfully converted.  */
 	  scm_lfwrite (encoded_output, output_len, port);
@@ -1034,11 +947,7 @@ display_string_using_iconv (const void *str, int narrow_p, size_t len,
 
 	      printed++;
 	    }
-	  else if (errno_save == E2BIG)
-            /* No space in output buffer for this input.  Keep
-               trucking.  */
-            continue;
-          else
+	  else
 	    /* Something bad happened that we can't handle: bail out.  */
 	    break;
 	}
@@ -1055,6 +964,8 @@ display_string_using_iconv (const void *str, int narrow_p, size_t len,
   return printed;
 }
 
+#undef STR_REF
+
 /* Display the LEN codepoints in STR to PORT according to STRATEGY;
    return the number of codepoints successfully displayed.  If NARROW_P,
    then STR is interpreted as a sequence of `char', denoting a Latin-1
@@ -1066,19 +977,24 @@ display_string (const void *str, int narrow_p,
 		scm_t_string_failed_conversion_handler strategy)
 
 {
-  scm_t_port_internal *pti;
+  scm_t_port *pt;
 
-  pti = SCM_PORT_GET_INTERNAL (port);
+  pt = SCM_PTAB_ENTRY (port);
 
-  if (pti->encoding_mode == SCM_PORT_ENCODING_MODE_UTF8)
+  if (pt->output_cd == (iconv_t) -1)
+    /* Initialize the conversion descriptors, if needed.  */
+    scm_i_set_port_encoding_x (port, pt->encoding);
+
+  /* FIXME: In 2.1, add a flag to determine whether a port is UTF-8.  */
+  if (pt->output_cd == (iconv_t) -1)
     return display_string_as_utf8 (str, narrow_p, len, port);
   else
     return display_string_using_iconv (str, narrow_p, len,
 				       port, strategy);
 }
 
-/* Attempt to display CH to PORT according to STRATEGY.  Return one if
-   CH was successfully displayed, zero otherwise (e.g., if it was not
+/* Attempt to display CH to PORT according to STRATEGY.  Return non-zero
+   if CH was successfully displayed, zero otherwise (e.g., if it was not
    representable in PORT's encoding.)  */
 static int
 display_character (scm_t_wchar ch, SCM port,
@@ -1087,34 +1003,6 @@ display_character (scm_t_wchar ch, SCM port,
   return display_string (&ch, 0, 1, port, strategy) == 1;
 }
 
-/* Same as 'display_string', but using the 'write' syntax.  */
-static size_t
-write_string (const void *str, int narrow_p,
-	      size_t len, SCM port,
-	      scm_t_string_failed_conversion_handler strategy)
-{
-  size_t printed;
-
-  printed = display_character ('"', port, strategy);
-
-  if (printed > 0)
-    {
-      size_t i;
-
-      for (i = 0; i < len; ++i)
-	{
-	  write_character (STR_REF (str, i), port, 1);
-	  printed++;
-	}
-
-      printed += display_character ('"', port, strategy);
-    }
-
-  return printed;
-}
-
-#undef STR_REF
-
 /* Attempt to pretty-print CH, a combining character, to PORT.  Return
    zero upon failure, non-zero otherwise.  The idea is to print CH above
    a dotted circle to make it more visible.  */
@@ -1228,7 +1116,7 @@ write_character (scm_t_wchar ch, SCM port, int string_escapes_p)
   int printed = 0;
   scm_t_string_failed_conversion_handler strategy;
 
-  strategy = PORT_CONVERSION_HANDLER (port);
+  strategy = scm_i_get_conversion_strategy (port);
 
   if (string_escapes_p)
     {
@@ -1276,29 +1164,6 @@ write_character (scm_t_wchar ch, SCM port, int string_escapes_p)
     write_character_escaped (ch, string_escapes_p, port);
 }
 
-/* Display STR to PORT from START inclusive to END exclusive.  */
-void
-scm_i_display_substring (SCM str, size_t start, size_t end, SCM port)
-{
-  int narrow_p;
-  const char *buf;
-  size_t len, printed;
-
-  buf = scm_i_string_data (str);
-  len = end - start;
-  narrow_p = scm_i_is_narrow_string (str);
-  buf += start * (narrow_p ? sizeof (char) : sizeof (scm_t_wchar));
-
-  printed = display_string (buf, narrow_p, end - start, port,
-			    PORT_CONVERSION_HANDLER (port));
-
-  if (SCM_UNLIKELY (printed < len))
-    scm_encoding_error (__func__, errno,
-			"cannot convert to output locale",
-			port, scm_c_string_ref (str, printed + start));
-}
-
-
 /* Print an integer.
  */
 
@@ -1485,7 +1350,7 @@ SCM_DEFINE (scm_simple_format, "simple-format", 2, 0, 1,
 	    "@var{message} can contain @code{~A} (was @code{%s}) and\n"
 	    "@code{~S} (was @code{%S}) escapes.  When printed,\n"
 	    "the escapes are replaced with corresponding members of\n"
-	    "@var{args}:\n"
+	    "@var{ARGS}:\n"
 	    "@code{~A} formats using @code{display} and @code{~S} formats\n"
 	    "using @code{write}.\n"
 	    "If @var{destination} is @code{#t}, then use the current output\n"
@@ -1501,7 +1366,6 @@ SCM_DEFINE (scm_simple_format, "simple-format", 2, 0, 1,
   if (scm_is_eq (destination, SCM_BOOL_T))
     {
       destination = port = scm_current_output_port ();
-      SCM_VALIDATE_OPORT_VALUE (1, destination);
     }
   else if (scm_is_false (destination))
     {
@@ -1605,7 +1469,7 @@ SCM_DEFINE (scm_write_char, "write-char", 1, 1, 0,
 
   port = SCM_COERCE_OUTPORT (port);
   if (!display_character (SCM_CHAR (chr), port,
-			  PORT_CONVERSION_HANDLER (port)))
+			  scm_i_get_conversion_strategy (port)))
     scm_encoding_error (__func__, errno,
 			"cannot convert to output locale",
 			port, chr);
@@ -1677,12 +1541,14 @@ SCM_DEFINE (scm_get_print_state, "get-print-state", 1, 0, 0,
 void
 scm_init_print ()
 {
-  SCM type;
+  SCM vtable, layout, type;
 
   scm_gc_register_root (&print_state_pool);
   scm_gc_register_root (&scm_print_state_vtable);
-  type = scm_make_vtable (scm_from_locale_string (SCM_PRINT_STATE_LAYOUT),
-                          SCM_BOOL_F);
+  vtable = scm_make_vtable_vtable (scm_nullstr, SCM_INUM0, SCM_EOL);
+  layout =
+    scm_make_struct_layout (scm_from_locale_string (SCM_PRINT_STATE_LAYOUT));
+  type = scm_make_struct (vtable, SCM_INUM0, scm_list_1 (layout));
   scm_set_struct_vtable_name_x (type, scm_from_latin1_symbol ("print-state"));
   scm_print_state_vtable = type;
 
