@@ -1,6 +1,6 @@
 ;;; -*- mode: scheme; coding: utf-8; -*-
 ;;;
-;;; Copyright (C) 2009, 2010, 2011 Free Software Foundation, Inc.
+;;; Copyright (C) 2009, 2010, 2011, 2012 Free Software Foundation, Inc.
 ;;;
 ;;; This library is free software; you can redistribute it and/or
 ;;; modify it under the terms of the GNU Lesser General Public
@@ -21,6 +21,7 @@
   #:use-module (srfi srfi-9)
   #:use-module (srfi srfi-9 gnu)
   #:use-module (srfi srfi-26)
+  #:use-module (ice-9 format)
 
   #:export (vlist? vlist-cons vlist-head vlist-tail vlist-null?
             vlist-null list->vlist vlist-ref vlist-drop vlist-take
@@ -66,9 +67,7 @@
 ;;;
 
 (define block-growth-factor
-  (let ((f (make-fluid)))
-    (fluid-set! f 2)
-    f))
+  (make-fluid 2))
 
 (define-syntax-rule (define-inline (name formals ...) body ...)
   ;; Work around the lack of an inliner.
@@ -547,23 +546,26 @@ with @var{equal?}."
 (define vhash-delq (cut vhash-delete <> <> eq? hashq))
 (define vhash-delv (cut vhash-delete <> <> eqv? hashv))
 
-(define (vhash-fold proc seed vhash)
-  "Fold over the key/pair elements of @var{vhash}.  For each pair call
-@var{proc} as @code{(@var{proc} key value result)}."
+(define (vhash-fold proc init vhash)
+  "Fold over the key/pair elements of @var{vhash} from left to right, with
+each call to @var{proc} having the form @code{(@var{proc} key value result)},
+where @var{result} is the result of the previous call to @var{proc} and
+@var{init} the value of @var{result} for the first call to @var{proc}."
   (vlist-fold (lambda (key+value result)
                 (proc (car key+value) (cdr key+value)
                       result))
-              seed
+              init
               vhash))
 
-(define (vhash-fold-right proc seed vhash)
-  "Fold over the key/pair elements of @var{vhash}, starting from the 0th
-element.  For each pair call @var{proc} as @code{(@var{proc} key value
-result)}."
+(define (vhash-fold-right proc init vhash)
+  "Fold over the key/pair elements of @var{vhash} from right to left, with
+each call to @var{proc} having the form @code{(@var{proc} key value result)},
+where @var{result} is the result of the previous call to @var{proc} and
+@var{init} the value of @var{result} for the first call to @var{proc}."
   (vlist-fold-right (lambda (key+value result)
                       (proc (car key+value) (cdr key+value)
                             result))
-                    seed
+                    init
                     vhash))
 
 (define* (alist->vhash alist #:optional (hash hash))
